@@ -1,13 +1,16 @@
 # SPDX-License-Identifier: MIT
 
+import collections
 from enum import Enum
 import errno
+import filecmp
 import os
 import re
 import shutil
 import subprocess
 import urllib.parse
 import urllib.request
+import stat
 import sys
 import tarfile
 import tempfile
@@ -121,7 +124,7 @@ def replace_at_vars(string, resolve):
 			raise RuntimeError("Unexpected substitution {}".format(varname))
 		return result
 
-	return re.sub(r'@(\w+)@', do_substitute, string)
+	return re.sub(r'@([\w:-]+)@', do_substitute, string)
 
 def installtree(src_root, dest_root):
 	for name in os.listdir(src_root):
@@ -144,6 +147,10 @@ def installtree(src_root, dest_root):
 			try_unlink(dest_path)
 			shutil.copy2(src_path, dest_path)
 
+class RollingIdUnavailableException(Exception):
+	def __init__(self, name):
+		super().__init__("No rolling_id specified for source {}".format(name))
+
 class ResetMode(Enum):
 	NONE = 0
 	RESET = 1
@@ -165,11 +172,14 @@ class ItemState:
 		self.updatable = updatable
 		self.timestamp = timestamp
 
+ArtifactFile = collections.namedtuple('ArtifactFile', ['name', 'filepath'])
+
 class Config:
 	def __init__(self, path):
 		self._config_path = path
 		self._root_yml = None
 		self._site_yml = dict()
+		self._commit_yml = dict()
 		self._sources = dict()
 		self._tool_pkgs = dict()
 		self._tool_stages = dict()
@@ -188,6 +198,13 @@ class Config:
 		except FileNotFoundError:
 			pass
 
+		commit_path = os.path.join(self._bootstrap_path, 'bootstrap-commits.yml')
+		try:
+			with open(commit_path, 'r') as f:
+				self._commit_yml = yaml.load(f, Loader=global_yaml_loader)
+		except FileNotFoundError:
+			pass
+
 		self._parse_yml(os.path.join(path, os.readlink(os.path.join(path, 'bootstrap.link'))),
 				self._root_yml)
 
@@ -265,6 +282,26 @@ class Config:
 				task = RunTask(self, task_yml)
 				self._tasks[task.name] = task
 
+	@property
+	def patch_author(self):
+		default = 'xbstrap'
+		if 'general' not in self._root_yml:
+			return default
+		return self._root_yml['general'].get('patch_author', default)
+
+	@property
+	def patch_email(self):
+		default = 'xbstrap@localhost'
+		if 'general' not in self._root_yml:
+			return default
+		return self._root_yml['general'].get('patch_email', default)
+
+	@property
+	def everything_by_default(self):
+		if 'general' not in self._root_yml:
+			return True
+		return self._root_yml['general'].get('everything_by_default', True)
+
 	@property
 	def repository_url(self):
 		if 'repository' not in self._root_yml:
@@ -329,6 +366,21 @@ class Config:
 		else:
 			return os.path.join(self.build_root, self._root_yml['directories']['packages'])
 
+	def get_option_value(self, name):
+		decl = None
+		for yml in self._root_yml.get('declare_options', []):
+			if yml['name'] == name:
+				assert not decl
+				decl = yml
+		if not decl:
+			raise KeyError()
+
+		defns = self._site_yml.get('define_options', dict())
+		if name in defns:
+			return defns[name]
+		else:
+			return decl.get('default', None)
+
 	def get_source(self, name):
 		return self._sources[name]
 
@@ -339,6 +391,9 @@ class Config:
 		if name in self._tasks:
 			return self._tasks[name]
 
+	def all_sources(self):
+		yield from self._sources.values()
+
 	def all_tools(self):
 		yield from self._tool_pkgs.values()
 
@@ -377,49 +432,53 @@ class ScriptStep:
 class RequirementsMixin:
 	@property
 	def source_dependencies(self):
-		# First, return automatic dependencies induced by 'sources_required'.
 		if 'sources_required' in self._this_yml:
 			yield from self._this_yml['sources_required']
 
-		# TODO: Secondly, return explicit dependencies.
-
-	@property
-	def tools_required(self):
-		if 'tools_required' in self._this_yml:
-			for yml in self._this_yml['tools_required']:
-				if isinstance(yml, dict):
-					if 'virtual' in yml:
-						continue
-					yield yml['tool']
-				else:
-					assert isinstance(yml, str)
-					yield yml
-
 	@property
 	def tool_dependencies(self):
-		# First, return automatic dependencies induced by 'tools_required'.
-		if 'tools_required' in self._this_yml:
-			for yml in self._this_yml['tools_required']:
-				if isinstance(yml, dict):
-					if 'virtual' in yml:
-						continue
-					if 'stage_dependencies' in yml:
-						for stage_name in yml['stage_dependencies']:
-							yield (yml['tool'], stage_name)
-					else:
-						yield (yml['tool'], None)
-				else:
-					assert isinstance(yml, str)
-					yield (yml, None)
+		deps = set(tool_name for tool_name, stage_name in self.tool_stage_dependencies)
+		yield from deps
 
-		# Secondly, return explicit dependencies.
-		if 'tool_dependencies' in self._this_yml:
-			for yml in self._this_yml['tool_dependencies']:
-				if isinstance(yml, dict):
-					yield (yml['tool'], yml['stage'])
+	@property
+	def tool_stage_dependencies(self):
+		tools_seen = set()
+		tools_stack = [ ]
+
+		def visit(stage):
+			if stage.subject_id in tools_seen:
+				return
+			tools_seen.add(stage.subject_id)
+			tools_stack.append(stage)
+
+		def visit_yml(yml):
+			if isinstance(yml, dict):
+				if 'virtual' in yml:
+					return
+				if 'stage_dependencies' in yml:
+					for stage_name in yml['stage_dependencies']:
+						visit(self._cfg.get_tool_pkg(yml['tool']).get_stage(stage_name))
 				else:
-					assert isinstance(yml, str)
-					yield (yml, None)
+					for stage in self._cfg.get_tool_pkg(yml['tool']).all_stages():
+						visit(stage)
+			else:
+				assert isinstance(yml, str)
+				for stage in self._cfg.get_tool_pkg(yml).all_stages():
+					visit(stage)
+
+		for yml in self._this_yml.get('tools_required', []):
+			visit_yml(yml)
+
+		while tools_stack:
+			stage = tools_stack.pop()
+			yield stage.subject_id
+
+			for yml in stage.pkg._this_yml.get('tools_required', []):
+				if not isinstance(yml, dict):
+					continue
+				if not yml.get('recursive', False):
+					continue
+				visit_yml(yml)
 
 	@property
 	def virtual_tools(self):
@@ -432,12 +491,9 @@ class RequirementsMixin:
 
 	@property
 	def pkg_dependencies(self):
-		# First, return automatic dependencies induced by 'pkgs_required'.
 		if 'pkgs_required' in self._this_yml:
 			yield from self._this_yml['pkgs_required']
 
-		# TODO: Secondly, return explicit dependencies.
-
 	@property
 	def task_dependencies(self):
 		if 'tasks_required' in self._this_yml:
@@ -504,6 +560,69 @@ class Source(RequirementsMixin):
 	def subject_type(self):
 		return 'source'
 
+	@property
+	def is_rolling_version(self):
+		return self._this_yml.get('rolling_version', False)
+
+	@property
+	def rolling_id(self):
+		commit_yml = self._cfg._commit_yml.get('commits', dict()).get(self._name, dict())
+		rolling_id = commit_yml.get('rolling_id')
+		if rolling_id is None:
+			raise RollingIdUnavailableException(self._name)
+		return rolling_id
+
+	def determine_rolling_id(self):
+		if 'git' in self._this_yml:
+			# Do some sanity checking: make sure that the repository is not shallow.
+			shallow_stdout = subprocess.check_output(['git', 'rev-parse', '--is-shallow-repository'],
+					cwd=self.source_dir).decode().strip()
+			if shallow_stdout == 'true':
+				raise RuntimeError("Cannot determine rolling version ID of source {} form shallow Git repository".format(self._name))
+			else:
+				assert shallow_stdout == 'false'
+
+			# Now count the number of commits.
+			if 'tag' in self._this_yml:
+				tracking_ref = 'refs/tags/' + self._this_yml['tag']
+			else:
+				ref = 'refs/heads/' + self._this_yml['branch']
+				tracking_ref = 'refs/remotes/origin/' + self._this_yml['branch']
+
+			try:
+				count_out = subprocess.check_output(['git', 'rev-list', '--count', tracking_ref],
+						cwd=self.source_dir).decode().strip()
+			except subprocess.CalledProcessError:
+				raise RuntimeError("Unable to determine rolling version ID of source {} via Git".format(self._name))
+			# Make sure that we get a valid number.
+			return str(int(count_out))
+		else:
+			raise RuntimeError('@ROLLING_ID@ requires git')
+
+	@property
+	def has_explicit_version(self):
+		return 'version' in self._this_yml
+
+	def compute_version(self, override_rolling_id=None):
+		if self.is_rolling_version:
+			if override_rolling_id is not None:
+				rolling_id = override_rolling_id
+			else:
+				rolling_id = self.rolling_id
+		else:
+			assert override_rolling_id is None
+
+		def substitute(varname):
+			if varname == 'ROLLING_ID':
+				assert self.is_rolling_version
+				return rolling_id
+
+		return replace_at_vars(self._this_yml.get('version', '0.0'), substitute)
+
+	@property
+	def version(self):
+		return self.compute_version()
+
 	@property
 	def sub_dir(self):
 		if 'subdir' in self._this_yml:
@@ -791,6 +910,10 @@ class HostPackage(RequirementsMixin):
 	def prefix_dir(self):
 		return os.path.join(self._cfg.tool_out_dir, self.name)
 
+	@property
+	def archive_file(self):
+		return os.path.join(self._cfg.tool_out_dir, self.name + '.tar.gz')
+
 	@property
 	def name(self):
 		return self._this_yml['name']
@@ -806,9 +929,13 @@ class HostPackage(RequirementsMixin):
 	@property
 	def is_default(self):
 		if 'default' not in self._this_yml:
-			return True
+			return self._cfg.everything_by_default
 		return self._this_yml['default']
 
+	@property
+	def stability_level(self):
+		return self._this_yml.get('stability_level', 'stable')
+
 	def all_stages(self):
 		yield from self._stages.values()
 
@@ -824,6 +951,23 @@ class HostPackage(RequirementsMixin):
 	def configure_steps(self):
 		yield from self._configure_steps
 
+	def compute_version(self, **kwargs):
+		source = self._cfg.get_source(self.source)
+
+		# If no version is specified, we fall back to 0.0_0.
+		if not source.has_explicit_version and 'revision' not in self._this_yml:
+			return source.compute_version(**kwargs) + '_0'
+
+		revision = self._this_yml.get('revision', 1)
+		if revision < 1:
+			raise RuntimeError("Tool {} specifies a revision < 1".format(self.name));
+
+		return source.compute_version(**kwargs) + '_' + str(revision)
+
+	@property
+	def version(self):
+		return self.compute_version()
+
 	def check_if_configured(self, settings):
 		path = os.path.join(self.build_dir, 'configured.xbstrap')
 		try:
@@ -845,6 +989,13 @@ class HostPackage(RequirementsMixin):
 				return ItemState(missing=True)
 		return ItemState()
 
+	def check_if_archived(self, settings):
+		try:
+			stat = os.stat(self.archive_file)
+		except FileNotFoundError:
+			return ItemState(missing=True)
+		return ItemState(timestamp=stat.st_mtime)
+
 class TargetPackage(RequirementsMixin):
 	def __init__(self, cfg, pkg_yml):
 		self._cfg = cfg
@@ -906,9 +1057,13 @@ class TargetPackage(RequirementsMixin):
 	@property
 	def is_default(self):
 		if 'default' not in self._this_yml:
-			return True
+			return self._cfg.everything_by_default
 		return self._this_yml['default']
 
+	@property
+	def stability_level(self):
+		return self._this_yml.get('stability_level', 'stable')
+
 	@property
 	def is_implicit(self):
 		if 'implict_package' not in self._this_yml:
@@ -923,9 +1078,22 @@ class TargetPackage(RequirementsMixin):
 	def build_steps(self):
 		yield from self._build_steps
 
+	def compute_version(self, **kwargs):
+		source = self._cfg.get_source(self.source)
+
+		# If no version is specified, we fall back to 0.0_0.
+		if not source.has_explicit_version and 'revision' not in self._this_yml:
+			return source.compute_version(**kwargs) + '_0'
+
+		revision = self._this_yml.get('revision', 1)
+		if revision < 1:
+			raise RuntimeError("Package {} specifies a revision < 1".format(self.name));
+
+		return source.compute_version(**kwargs) + '_' + str(revision)
+
 	@property
 	def version(self):
-		return "0.0_0"
+		return self.compute_version()
 
 	def get_task(self, task):
 		if task in self._tasks:
@@ -1058,6 +1226,23 @@ class RunTask(RequirementsMixin):
 	def subject_type(self):
 		return 'task'
 
+	@property
+	def artifact_files(self):
+		def substitute(varname):
+			if varname == 'SOURCE_ROOT':
+				return self._cfg.source_root
+			elif varname == 'BUILD_ROOT':
+				return self._cfg.build_root
+			elif varname == 'SYSROOT_DIR':
+				return self._cfg.sysroot_dir
+			elif varname.startswith('OPTION:'):
+				return cfg.get_option_value(varname[7:])
+
+		entries = self._this_yml.get('artifact_files', [])
+		for e in entries:
+			path = replace_at_vars(e['path'], substitute)
+			yield ArtifactFile(e['name'], os.path.join(path, e['name']))
+
 def config_for_dir():
 	return Config('')
 
@@ -1068,14 +1253,14 @@ class EnvironmentComposer:
 		self.shared_lib_dirs = [ ]
 		self.aclocal_dirs = [ ]
 
-	def compose(self, for_package=False, use_target_pkgconfig=False):
+	def compose(self, for_package=False, explicit_pkgconfig=False):
 		environ = os.environ.copy()
 
 		environ['XBSTRAP_SOURCE_ROOT'] = self.cfg.source_root
 		environ['XBSTRAP_BUILD_ROOT'] = self.cfg.build_root
 		environ['XBSTRAP_SYSROOT_DIR'] = self.cfg.sysroot_dir
 
-		if for_package and not use_target_pkgconfig:
+		if for_package and not explicit_pkgconfig:
 			pkgcfg_libdir = os.path.join(self.cfg.sysroot_dir, 'usr', 'lib', 'pkgconfig')
 			pkgcfg_libdir += ':' + os.path.join(self.cfg.sysroot_dir, 'usr', 'share', 'pkgconfig')
 
@@ -1111,25 +1296,6 @@ def run_tool(cfg, args, tool_pkgs=[], virtual_tools=[], workdir=None, extra_envi
 	pkg_queue = []
 	pkg_visited = set()
 
-	# /bin directory for virtual tools.
-	use_target_pkgconfig = False
-	vb = tempfile.TemporaryDirectory()
-
-	for yml in virtual_tools:
-		if yml['virtual'] == 'pkgconfig-for-target':
-			vscript = os.path.join(vb.name, '{}-pkg-config'.format(yml['triple']))
-			with open(vscript, 'wt') as f:
-				f.write('#!/bin/sh\n'
-					+ 'PKG_CONFIG_PATH= '
-					+ ' PKG_CONFIG_SYSROOT_DIR="${XBSTRAP_SYSROOT_DIR}"'
-					+ ' PKG_CONFIG_LIBDIR="${XBSTRAP_SYSROOT_DIR}/usr/lib/pkgconfig'
-					+ ':${XBSTRAP_SYSROOT_DIR}/usr/share/pkgconfig"'
-					+ ' pkg-config $@\n')
-			os.chmod(vscript, 0o775)
-			use_targe_pkgconfig = True
-		else:
-			raise RuntimeError("Unknown virtual tool {}".format(yml['virtual']))
-
 	for pkg in tool_pkgs:
 		assert pkg.name not in pkg_visited
 		pkg_queue.append(pkg)
@@ -1146,6 +1312,37 @@ def run_tool(cfg, args, tool_pkgs=[], virtual_tools=[], workdir=None, extra_envi
 			pkg_visited.add(dep_name)
 		i += 1
 
+	# /bin directory for virtual tools.
+	explicit_pkgconfig = False
+	vb = tempfile.TemporaryDirectory()
+
+	for yml in virtual_tools:
+		if yml['virtual'] == 'pkgconfig-for-host':
+			vscript = os.path.join(vb.name, yml['program_name'])
+			paths = []
+			for tool in pkg_queue:
+				paths.append(tool.prefix_dir + '/lib/pkgconfig')
+				paths.append(tool.prefix_dir + '/share/pkgconfig')
+			with open(vscript, 'wt') as f:
+				f.write('#!/bin/sh\n'
+					+ 'PKG_CONFIG_PATH=' + ':'.join(paths)
+					+ ' pkg-config $@\n')
+			os.chmod(vscript, 0o775)
+			explicit_pkgconfig = True
+		elif yml['virtual'] == 'pkgconfig-for-target':
+			vscript = os.path.join(vb.name, '{}-pkg-config'.format(yml['triple']))
+			with open(vscript, 'wt') as f:
+				f.write('#!/bin/sh\n'
+					+ 'PKG_CONFIG_PATH= '
+					+ ' PKG_CONFIG_SYSROOT_DIR="${XBSTRAP_SYSROOT_DIR}"'
+					+ ' PKG_CONFIG_LIBDIR="${XBSTRAP_SYSROOT_DIR}/usr/lib/pkgconfig'
+					+ ':${XBSTRAP_SYSROOT_DIR}/usr/share/pkgconfig"'
+					+ ' pkg-config $@\n')
+			os.chmod(vscript, 0o775)
+			explicit_pkgconfig = True
+		else:
+			raise RuntimeError("Unknown virtual tool {}".format(yml['virtual']))
+
 	composer = EnvironmentComposer(cfg)
 	composer.path_dirs.append(vb.name)
 	for pkg in pkg_queue:
@@ -1154,7 +1351,7 @@ def run_tool(cfg, args, tool_pkgs=[], virtual_tools=[], workdir=None, extra_envi
 			composer.shared_lib_dirs.append(os.path.join(pkg.prefix_dir, 'lib'))
 		if pkg.exports_aclocal:
 			composer.aclocal_dirs.append(os.path.join(pkg.prefix_dir, 'share/aclocal'))
-	environ = composer.compose(for_package=for_package, use_target_pkgconfig=use_target_pkgconfig)
+	environ = composer.compose(for_package=for_package, explicit_pkgconfig=explicit_pkgconfig)
 	environ.update(extra_environ)
 
 	output = None # Default: Do not redirect output.
@@ -1222,6 +1419,10 @@ def fetch_src(cfg, src):
 					cwd=src.source_dir)
 
 		shallow = not source.get('disable_shallow_fetch', False)
+		# We have to disable shallow fetches to get rolling versions right.
+		if src.is_rolling_version:
+			shallow = False
+
 		args = ['git', 'fetch']
 		if 'tag' in source:
 			if shallow:
@@ -1318,7 +1519,10 @@ def checkout_src(cfg, src, settings):
 			with tarfile.open(src.source_archive_file,
 					'r|' + compression[src.source_archive_format]) as tar:
 				for info in tar:
-					prefix = source['extract_path'] + '/'
+					if 'extract_path' not in source:
+						prefix = ''
+					else:
+						prefix = source['extract_path'] + '/'
 					if info.name.startswith(prefix):
 						info.name = src.name + '/' + info.name[len(prefix):]
 						tar.extract(info, src.sub_dir)
@@ -1341,8 +1545,12 @@ def patch_src(cfg, src):
 		if not patch.endswith('.patch'):
 			continue
 		if 'git' in source:
-			subprocess.check_call(['git', 'am', os.path.join(src.patch_dir, patch)],
-				cwd=src.source_dir)
+			environ = os.environ.copy()
+			environ['GIT_COMMITTER_NAME'] = cfg.patch_author
+			environ['GIT_COMMITTER_EMAIL'] = cfg.patch_email
+			subprocess.check_call(['git', 'am', '-3', '--no-gpg-sign', '--committer-date-is-author-date',
+					os.path.join(src.patch_dir, patch)],
+				env=environ, cwd=src.source_dir)
 		elif 'hg' in source:
 			subprocess.check_call(['hg', 'import', os.path.join(src.patch_dir, patch)],
 				cwd=src.source_dir)
@@ -1356,7 +1564,7 @@ def patch_src(cfg, src):
 def regenerate_src(cfg, src):
 	for step in src.regenerate_steps:
 		tool_pkgs = []
-		for dep_name in src.tools_required:
+		for dep_name in src.tool_dependencies:
 			tool_pkgs.append(cfg.get_tool_pkg(dep_name))
 
 		def substitute(varname):
@@ -1366,8 +1574,10 @@ def regenerate_src(cfg, src):
 				return cfg.build_root
 			elif varname == 'SYSROOT_DIR':
 				return cfg.sysroot_dir
-			if varname == 'THIS_SOURCE_DIR':
+			elif varname == 'THIS_SOURCE_DIR':
 				return src.source_dir
+			elif varname.startswith('OPTION:'):
+				return cfg.get_option_value(varname[7:])
 
 		run_step(cfg, step, src.source_dir, substitute, tool_pkgs, src.virtual_tools)
 
@@ -1385,7 +1595,7 @@ def configure_tool(cfg, pkg):
 
 	for step in pkg.configure_steps:
 		tool_pkgs = []
-		for dep_name in pkg.tools_required:
+		for dep_name in pkg.tool_dependencies:
 			tool_pkgs.append(cfg.get_tool_pkg(dep_name))
 
 		def substitute(varname):
@@ -1404,6 +1614,8 @@ def configure_tool(cfg, pkg):
 			elif varname == 'PARALLELISM':
 				nthreads = get_concurrency()
 				return str(nthreads)
+			elif varname.startswith('OPTION:'):
+				return cfg.get_option_value(varname[7:])
 
 		run_step(cfg, step, pkg.build_dir, substitute, tool_pkgs, pkg.virtual_tools)
 
@@ -1415,7 +1627,7 @@ def compile_tool_stage(cfg, stage):
 
 	for step in stage.compile_steps:
 		tool_pkgs = []
-		for dep_name in pkg.tools_required:
+		for dep_name in pkg.tool_dependencies:
 			tool_pkgs.append(cfg.get_tool_pkg(dep_name))
 
 		def substitute(varname):
@@ -1434,6 +1646,8 @@ def compile_tool_stage(cfg, stage):
 			elif varname == 'PARALLELISM':
 				nthreads = get_concurrency()
 				return str(nthreads)
+			elif varname.startswith('OPTION:'):
+				return cfg.get_option_value(varname[7:])
 
 		run_step(cfg, step, pkg.build_dir, substitute, tool_pkgs, pkg.virtual_tools)
 
@@ -1449,7 +1663,7 @@ def install_tool_stage(cfg, stage):
 
 	for step in stage.install_steps:
 		tool_pkgs = []
-		for dep_name in pkg.tools_required:
+		for dep_name in pkg.tool_dependencies:
 			tool_pkgs.append(cfg.get_tool_pkg(dep_name))
 
 		def substitute(varname):
@@ -1465,11 +1679,18 @@ def install_tool_stage(cfg, stage):
 				return pkg.build_dir
 			elif varname == 'PREFIX':
 				return pkg.prefix_dir
+			elif varname.startswith('OPTION:'):
+				return cfg.get_option_value(varname[7:])
 
 		run_step(cfg, step, pkg.build_dir, substitute, tool_pkgs, pkg.virtual_tools)
 
 	stage.mark_as_installed()
 
+def archive_tool(cfg, tool):
+	with tarfile.open(tool.archive_file, 'w|gz') as tar:
+		for ent in os.listdir(tool.prefix_dir):
+			tar.add(os.path.join(tool.prefix_dir, ent), arcname=ent)
+
 # ---------------------------------------------------------------------------------------
 # Package building.
 # ---------------------------------------------------------------------------------------
@@ -1482,7 +1703,7 @@ def configure_pkg(cfg, pkg):
 
 	for step in pkg.configure_steps:
 		tool_pkgs = []
-		for dep_name in pkg.tools_required:
+		for dep_name in pkg.tool_dependencies:
 			tool_pkgs.append(cfg.get_tool_pkg(dep_name))
 
 		def substitute(varname):
@@ -1496,13 +1717,20 @@ def configure_pkg(cfg, pkg):
 				return src.source_dir
 			elif varname == 'THIS_BUILD_DIR':
 				return pkg.build_dir
+			elif varname == 'PARALLELISM':
+				nthreads = get_concurrency()
+				return str(nthreads)
+			elif varname.startswith('OPTION:'):
+				return cfg.get_option_value(varname[7:])
+			elif varname.startswith('OPTION:'):
+				return cfg.get_option_value(varname[7:])
 
 		run_step(cfg, step, pkg.build_dir, substitute, tool_pkgs, pkg.virtual_tools,
 				for_package=True)
 
 	pkg.mark_as_configured()
 
-def build_pkg(cfg, pkg):
+def build_pkg(cfg, pkg, reproduce=False):
 	src = cfg.get_source(pkg.source)
 
 	try_mkdir(cfg.package_out_dir)
@@ -1511,7 +1739,7 @@ def build_pkg(cfg, pkg):
 
 	for step in pkg.build_steps:
 		tool_pkgs = []
-		for dep_name in pkg.tools_required:
+		for dep_name in pkg.tool_dependencies:
 			tool_pkgs.append(cfg.get_tool_pkg(dep_name))
 
 		def substitute(varname):
@@ -1530,16 +1758,86 @@ def build_pkg(cfg, pkg):
 			elif varname == 'PARALLELISM':
 				nthreads = get_concurrency()
 				return str(nthreads)
+			elif varname.startswith('OPTION:'):
+				return cfg.get_option_value(varname[7:])
 
 		run_step(cfg, step, pkg.build_dir, substitute, tool_pkgs, pkg.virtual_tools,
 				for_package=True)
 
 	postprocess_libtool(cfg, pkg)
 
-	try_rmtree(pkg.staging_dir)
-	os.rename(pkg.collect_dir, pkg.staging_dir)
+	if not reproduce:
+		try_rmtree(pkg.staging_dir)
+		os.rename(pkg.collect_dir, pkg.staging_dir)
+	else:
+		def discover_dirtree(root):
+			s = set()
+
+			def recurse(subdir=''):
+				for dent in os.scandir(os.path.join(root, subdir)):
+					path = os.path.join(subdir, dent.name)
+					s.add(path)
+					if dent.is_dir(follow_symlinks=False):
+						recurse(path)
+
+			recurse()
+			return s
+
+		repro_paths = discover_dirtree(pkg.collect_dir)
+		exist_paths = discover_dirtree(pkg.staging_dir)
+
+		repro_only = repro_paths.difference(exist_paths)
+		exist_only = exist_paths.difference(repro_paths)
+		if repro_only:
+			raise RuntimeError("Paths {} only exist in reproducted build".format(
+					', '.join(repro_only)))
+		if exist_only:
+			raise RuntimeError("Paths {} only exist in existing build".format(
+					', '.join(repro_only)))
+
+		any_issues = False
+		for path in repro_paths:
+			repro_stat = os.stat(os.path.join(pkg.collect_dir, path))
+			exist_stat = os.stat(os.path.join(pkg.collect_dir, path))
+
+			if stat.S_IFMT(repro_stat.st_mode) != stat.S_IFMT(exist_stat.st_mode):
+				print("{}xbstrap{}: File type mismatch in file {}".format(
+						colorama.Style.BRIGHT, colorama.Style.RESET_ALL,
+						path))
+				any_issues = True
+				continue
+
+			if stat.S_ISREG(repro_stat.st_mode):
+				if not filecmp.cmp(os.path.join(pkg.collect_dir, path),
+						os.path.join(pkg.staging_dir, path),
+						shallow=False):
+					print("{}xbstrap{}: Content mismatch in file {}".format(
+							colorama.Style.BRIGHT, colorama.Style.RESET_ALL,
+							path))
+					any_issues = True
+					continue
+
+		if not any_issues:
+			print("{}xbstrap{}: Build was reproduced exactly".format(
+					colorama.Style.BRIGHT, colorama.Style.RESET_ALL))
+		else:
+			raise RuntimeError('Could not reproduce all files')
+
+def pack_pkg(cfg, pkg, reproduce=False):
+	# Sanity checking: make sure that the rolling ID matches the expected one.
+	src = cfg.get_source(pkg.source)
+	if src.is_rolling_version:
+		actual_rolling_id = src.determine_rolling_id()
+		try:
+			if src.rolling_id != actual_rolling_id:
+				raise RuntimeError("Rolling ID of package {} does not match true rolling ID".format(pkg.name))
+		except RollingIdUnavailableException:
+			pass
+	else:
+		actual_rolling_id = None
+
+	version = pkg.compute_version(override_rolling_id=actual_rolling_id)
 
-def pack_pkg(cfg, pkg):
 	if cfg.use_xbps:
 		try_mkdir(cfg.xbps_repository_dir)
 
@@ -1549,23 +1847,41 @@ def pack_pkg(cfg, pkg):
 
 		args = ['xbps-create', '-A', 'x86_64',
 			'-s', pkg.name,
-			'-n', '{}-{}'.format(pkg.name, pkg.version),
+			'-n', '{}-{}'.format(pkg.name, version),
 			'-D', pkg.xbps_dependency_string(),
 			pkg.staging_dir
 		]
-		print("{}xbstrap{}: Running {}".format(
-				colorama.Style.BRIGHT, colorama.Style.RESET_ALL,
-				args))
-		subprocess.call(args, cwd=cfg.xbps_repository_dir, stdout=output)
 
-		args = ['xbps-rindex', '-fa',
-				os.path.join(cfg.xbps_repository_dir,
-				'{}-{}.x86_64.xbps'.format(pkg.name, pkg.version))
-		]
-		print("{}xbstrap{}: Running {}".format(
-				colorama.Style.BRIGHT, colorama.Style.RESET_ALL,
-				args))
-		subprocess.call(args, stdout=output)
+		xbps_file = '{}-{}.x86_64.xbps'.format(pkg.name, version)
+
+		if not reproduce:
+			print("{}xbstrap{}: Running {}".format(
+					colorama.Style.BRIGHT, colorama.Style.RESET_ALL,
+					args))
+			subprocess.call(args, cwd=cfg.xbps_repository_dir, stdout=output)
+
+			args = ['xbps-rindex', '-fa',
+					os.path.join(cfg.xbps_repository_dir, xbps_file)
+			]
+			print("{}xbstrap{}: Running {}".format(
+					colorama.Style.BRIGHT, colorama.Style.RESET_ALL,
+					args))
+			subprocess.call(args, stdout=output)
+		else:
+			print("{}xbstrap{}: Running {}".format(
+					colorama.Style.BRIGHT, colorama.Style.RESET_ALL,
+					args))
+			subprocess.call(args, cwd=cfg.package_out_dir, stdout=output)
+
+			if not filecmp.cmp(os.path.join(cfg.package_out_dir, xbps_file),
+					os.path.join(cfg.xbps_repository_dir, xbps_file),
+					shallow=False):
+				print("{}xbstrap{}: Mismatch in {}".format(
+						colorama.Style.BRIGHT, colorama.Style.RESET_ALL, xbps_file))
+				raise RuntimeError('Could not reproduce pack')
+
+			print("{}xbstrap{}: Pack was reproduced exactly".format(
+					colorama.Style.BRIGHT, colorama.Style.RESET_ALL))
 	else:
 		raise RuntimeError('Package management configuration does not support pack')
 
@@ -1598,7 +1914,7 @@ def archive_pkg(cfg, pkg):
 
 def run_task(cfg, task):
 	tools_required = []
-	for dep_name in task.tools_required:
+	for dep_name in task.tool_dependencies:
 		tools_required.append(cfg.get_tool_pkg(dep_name))
 
 	def substitute(varname):
@@ -1611,6 +1927,8 @@ def run_task(cfg, task):
 		elif varname == 'PARALLELISM':
 			nthreads = get_concurrency()
 			return str(nthreads)
+		elif varname.startswith('OPTION:'):
+			return cfg.get_option_value(varname[7:])
 
 	run_step(cfg, task.script_step, cfg.source_root, substitute, tools_required, task.virtual_tools,
 			for_package=False)
@@ -1620,7 +1938,7 @@ def run_pkg_task(cfg, task):
 	src = cfg.get_source(task.pkg.source)
 
 	tools_required = []
-	for dep_name in task.pkg.tools_required:
+	for dep_name in task.pkg.tool_dependencies:
 		tools_required.append(cfg.get_tool_pkg(dep_name))
 
 	def substitute(varname):
@@ -1639,6 +1957,8 @@ def run_pkg_task(cfg, task):
 		elif varname == 'PARALLELISM':
 			nthreads = get_concurrency()
 			return str(nthreads)
+		elif varname.startswith('OPTION:'):
+			return cfg.get_option_value(varname[7:])
 
 	run_step(cfg, task.script_step, task.pkg.build_dir, substitute, tools_required,
 			task.pkg.virtual_tools, for_package=False)
@@ -1647,7 +1967,7 @@ def run_tool_task(cfg, task):
 	src = cfg.get_source(task.pkg.source)
 
 	tools_required = []
-	for dep_name in task.pkg.tools_required:
+	for dep_name in task.pkg.tool_dependencies:
 		tools_required.append(cfg.get_tool_pkg(dep_name))
 
 	def substitute(varname):
@@ -1666,6 +1986,8 @@ def run_tool_task(cfg, task):
 		elif varname == 'PARALLELISM':
 			nthreads = get_concurrency()
 			return str(nthreads)
+		elif varname.startswith('OPTION:'):
+			return cfg.get_option_value(varname[7:])
 
 	run_step(cfg, task.script_step, task.pkg.build_dir, substitute, tools_required,
 			task.pkg.virtual_tools, for_package=False)
@@ -1688,14 +2010,17 @@ class Action(Enum):
 	# Package-related actions.
 	CONFIGURE_PKG = 8
 	BUILD_PKG = 9
-	PACK_PKG = 10
-	INSTALL_PKG = 11
-	ARCHIVE_PKG = 12
-	RUN = 13
-	RUN_PKG = 14
-	RUN_TOOL = 15
-	WANT_TOOL = 16
-	WANT_PKG = 17
+	REPRODUCE_BUILD_PKG = 10
+	PACK_PKG = 11
+	REPRODUCE_PACK_PKG = 12
+	INSTALL_PKG = 13
+	ARCHIVE_TOOL = 14
+	ARCHIVE_PKG = 15
+	RUN = 16
+	RUN_PKG = 17
+	RUN_TOOL = 18
+	WANT_TOOL = 19
+	WANT_PKG = 20
 
 Action.strings = {
 	Action.FETCH_SRC: 'fetch',
@@ -1707,8 +2032,11 @@ Action.strings = {
 	Action.INSTALL_TOOL_STAGE: 'install-tool',
 	Action.CONFIGURE_PKG: 'configure',
 	Action.BUILD_PKG: 'build',
+	Action.REPRODUCE_BUILD_PKG: 'reproduce-build',
 	Action.PACK_PKG: 'pack',
+	Action.REPRODUCE_PACK_PKG: 'reproduce-pack',
 	Action.INSTALL_PKG: 'install',
+	Action.ARCHIVE_TOOL: 'archive-tool',
 	Action.ARCHIVE_PKG: 'archive',
 	Action.RUN: 'run',
 	Action.RUN_PKG: 'run',
@@ -1778,8 +2106,11 @@ class PlanItem:
 			Action.INSTALL_TOOL_STAGE: lambda s, c: s.check_if_installed(c),
 			Action.CONFIGURE_PKG: lambda s, c: s.check_if_configured(c),
 			Action.BUILD_PKG: lambda s, c: s.check_staging(c),
+			Action.REPRODUCE_BUILD_PKG: lambda s, c: ItemState(missing=True),
 			Action.PACK_PKG: lambda s, c: s.check_if_packed(c),
+			Action.REPRODUCE_PACK_PKG: lambda s, c: ItemState(missing=True),
 			Action.INSTALL_PKG: lambda s, c: s.check_if_installed(c),
+			Action.ARCHIVE_TOOL: lambda s, c: s.check_if_archived(c),
 			Action.ARCHIVE_PKG: lambda s, c: ItemState(missing=True),
 			Action.RUN: lambda s, c: ItemState(missing=True),
 			Action.RUN_PKG: lambda s, c: ItemState(missing=True),
@@ -1820,6 +2151,7 @@ class Plan:
 		self.hard = False
 		self.only_wanted = False
 		self.keep_going = False
+		self.progress_file = None
 
 	@property
 	def cfg(self):
@@ -1840,13 +2172,10 @@ class Plan:
 				item.require_edges.add((action.PATCH_SRC, dep_source))
 
 		def add_tool_dependencies(s):
-			for (tool_name, stage_name) in s.tool_dependencies:
+			for (tool_name, stage_name) in s.tool_stage_dependencies:
 				dep_tool = self._cfg.get_tool_pkg(tool_name)
 				if self.build_scope is not None and dep_tool not in self.build_scope:
 					item.require_edges.add((action.WANT_TOOL, dep_tool))
-				elif stage_name is None:
-					item.require_edges.update([(action.INSTALL_TOOL_STAGE, stage)
-							for stage in dep_tool.all_stages()])
 				else:
 					tool_stage = dep_tool.get_stage(stage_name)
 					item.require_edges.add((action.INSTALL_TOOL_STAGE, tool_stage))
@@ -1886,6 +2215,7 @@ class Plan:
 			src = self._cfg.get_source(subject.source)
 			item.build_edges.add((action.REGENERATE_SRC, src))
 
+			add_source_dependencies(subject)
 			add_tool_dependencies(subject)
 			add_pkg_dependencies(subject)
 			add_task_dependencies(subject)
@@ -1893,6 +2223,7 @@ class Plan:
 		elif action == Action.COMPILE_TOOL_STAGE:
 			item.build_edges.add((action.CONFIGURE_TOOL, subject.pkg))
 
+			add_source_dependencies(subject)
 			add_tool_dependencies(subject.pkg)
 			add_tool_dependencies(subject)
 			add_pkg_dependencies(subject)
@@ -1911,22 +2242,24 @@ class Plan:
 			item.build_edges.add((action.REGENERATE_SRC, src))
 
 			# Configuration requires all dependencies to be present.
+			add_source_dependencies(subject)
 			add_implicit_pkgs()
 			add_pkg_dependencies(subject)
 			add_tool_dependencies(subject)
 			add_task_dependencies(subject)
 
-		elif action == Action.BUILD_PKG:
+		elif action == Action.BUILD_PKG or action == Action.REPRODUCE_BUILD_PKG:
 			item.build_edges.add((action.CONFIGURE_PKG, subject))
 
 			# Usually dependencies will already be installed during the configuration phase.
 			# However, if the sysroot is removed, we might need to install again.
+			add_source_dependencies(subject)
 			add_implicit_pkgs()
 			add_pkg_dependencies(subject)
 			add_tool_dependencies(subject)
 			add_task_dependencies(subject)
 
-		elif action == Action.PACK_PKG:
+		elif action == Action.PACK_PKG or action == Action.REPRODUCE_PACK_PKG:
 			item.build_edges.add((action.BUILD_PKG, subject))
 
 		elif action == Action.INSTALL_PKG:
@@ -1941,6 +2274,10 @@ class Plan:
 			add_implicit_pkgs()
 			add_pkg_dependencies(subject)
 
+		elif action == Action.ARCHIVE_TOOL:
+			for stage in subject.all_stages():
+				item.build_edges.add((action.INSTALL_TOOL_STAGE, stage))
+
 		elif action == Action.ARCHIVE_PKG:
 			item.build_edges.add((action.BUILD_PKG, subject))
 
@@ -2187,12 +2524,29 @@ class Plan:
 				if edge_item.exec_status != ExecutionStatus.SUCCESS:
 					any_failed_edges = True
 
+			def emit_progress(status):
+				if self.progress_file is not None:
+					yml = {
+						'n_this': n + 1,
+						'n_all': len(scheduled),
+						'status': status,
+						'action': Action.strings[action],
+						'subject': subject.subject_id,
+						'artifact_files': []
+					}
+					if action == Action.RUN:
+						for af in subject.artifact_files:
+							yml['artifact_files'].append({'name': af.name, 'filepath': af.filepath})
+					self.progress_file.write(yaml.safe_dump(yml, explicit_end=True))
+					self.progress_file.flush()
+
 			if self.keep_going and any_failed_edges:
 				print('{}xbstrap{}: Skipping action {} of {} due to failed prerequisites [{}/{}]'.format(
 						colorama.Style.BRIGHT, colorama.Style.RESET_ALL,
 						Action.strings[action], subject.subject_id,
 						n + 1, len(scheduled)))
 				item.exec_status = ExecutionStatus.PREREQS_FAILED
+				emit_progress('prereqs-failed')
 				any_failed_items = True
 				continue
 
@@ -2200,6 +2554,7 @@ class Plan:
 				if not self.keep_going:
 					raise ExecutionFailureException(action, subject)
 				item.exec_status = ExecutionStatus.NOT_WANTED
+				emit_progress('not-wanted')
 				any_failed_items = True
 				continue
 
@@ -2227,10 +2582,16 @@ class Plan:
 					configure_pkg(self._cfg, subject)
 				elif action == Action.BUILD_PKG:
 					build_pkg(self._cfg, subject)
+				elif action == Action.REPRODUCE_BUILD_PKG:
+					build_pkg(self._cfg, subject, reproduce=True)
 				elif action == Action.PACK_PKG:
 					pack_pkg(self._cfg, subject)
+				elif action == Action.REPRODUCE_PACK_PKG:
+					pack_pkg(self._cfg, subject, reproduce=True)
 				elif action == Action.INSTALL_PKG:
 					install_pkg(self._cfg, subject)
+				elif action == Action.ARCHIVE_TOOL:
+					archive_tool(self._cfg, subject)
 				elif action == Action.ARCHIVE_PKG:
 					archive_pkg(self._cfg, subject)
 				elif action == Action.RUN:
@@ -2250,10 +2611,12 @@ class Plan:
 				else:
 					raise AssertionError("Unexpected action")
 				item.exec_status = ExecutionStatus.SUCCESS
+				emit_progress('success')
 			except (subprocess.CalledProcessError, ExecutionFailureException):
+				item.exec_status = ExecutionStatus.STEP_FAILED
+				emit_progress('failure')
 				if not self.keep_going:
 					raise ExecutionFailureException(action, subject)
-				item.exec_status = ExecutionStatus.STEP_FAILED
 				any_failed_items = True
 
 		if any_failed_items:
